param (
    [String] $Secret = $null,
    [String] $WebState = $null,
    [Switch] $help = $false,
    [Switch] $ActiveMQ = $false,
    [Switch] $All = $false,
    [Switch] $Base = $false,
    [Switch] $Everything = $false,
    [Switch] $Gov = $false,
    [Switch] $Governance = $false,
    [Switch] $IDGov = $false,
    [Switch] $Java = $false,
    [Switch] $JDK = $false,
    [Switch] $JRE = $false,
    [Switch] $Mail = $false,
    [Switch] $PG = $false,
    [Switch] $Postgres = $false,
    [Switch] $PostgreSQL = $false,
    [Switch] $Tomcat = $false,
    [Switch] $Web = $false
)

Add-Type -AssemblyName System.IO.Compression.FileSystem

#Versions
$VSN_ACTIVEMQ           = "5.15.9"
$VSN_JDK                = "8.40.0.25-ca-jdk8.0.222"
$VSN_POSTGRESQL         = "11.5-1"
$VSN_POSTGRESQL_SERVICE = "11"
$VSN_TOMCAT             = "9.0.22"

#Names
$NAME_ACTIVEMQ          = "activemq"
$NAME_POSTGRES          = "postgres"
$NAME_TOMCAT            = "tomcat"
$NAME_TOMCAT_LAUNCH     = "start-tomcat.bat"
$NAME_TOMCAT_QUIT       = "stop-tomcat.bat"
$NAME_TOMCAT_SERVICE    = "TomcatService.exe"

#Directories
$DIR_ACTIVEMQ           = "${NAME_ACTIVEMQ}"
$DIR_BACKUP             = "backup"
$DIR_BASE               = "c:\netiq\idm\apps"
$DIR_COMPRESSED         = "${PSScriptRoot}\compressed"
$DIR_JDK                = "jdk"
$DIR_LOG_BASE           = "${PSScriptRoot}"
$DIR_LOG_PARENT         = "logs"
$DIR_LOG                = "${DIR_LOG_BASE}\${DIR_LOG_PARENT}"
$DIR_POSTGRES           = "${NAME_POSTGRES}"
$DIR_POSTGRES_DATA      = "data"
$DIR_PRISTINE           = "${PSScriptRoot}\untouched"
$DIR_TOMCAT             = "${NAME_TOMCAT}"

#Files
$FILE_ACTIVEMQ          = "${DIR_COMPRESSED}\apache-activemq-${VSN_ACTIVEMQ}-bin.zip"
$FILE_JDK               = "${DIR_COMPRESSED}\zulu${VSN_JDK}-win_x64.zip"
$FILE_POSTGRESQL        = "${DIR_COMPRESSED}\postgresql-${VSN_POSTGRESQL}-windows-x64.exe"
$FILE_LOG               = "${DIR_LOG}\IDGov-Core-Helper.log"
$FILE_TOMCAT            = "${DIR_COMPRESSED}\apache-tomcat-${VSN_TOMCAT}-windows-x64.zip"

#Services (final state can be `start`, `stop`, and `restart`)
$SERV_TOMCAT_SERVICE    = "IDM Apps Tomcat Service"
$SERV_TOMCAT_STATE      = ""

#Parameters
$PARAM_OSP_LDAP_TYPE    = "edir"
$PARAM_OSP_LOGGING      = "WARN"
$PARAM_POSTGRES_PORT    = "5432"
$PARAM_TOMCAT_SERVER    = "localhost"
$PARAM_TOMCAT_OPTS_COMMON_ALL = "-Dnetiqtomcatcustomsrvidentify=netiqtomcatcustomsrvidentify;-Djava.awt.headless=true;-Dcom.netiq.ism.config=${DIR_BASE}\${DIR_TOMCAT}\conf\ism-configuration.properties;"
$PARAM_TOMCAT_OPTS_OSP  = "-Dcom.netiq.osp.ext-context-file=${DIR_BASE}\osp\lib\osp-conf-${PARAM_OSP_LDAP_TYPE}.jar;-Dcom.netiq.idm.osp.logging.level=${PARAM_OSP_LOGGING};-Dcom.netiq.idm.osp.client.host=${PARAM_TOMCAT_SERVER};-Dcom.netiq.idm.osp.logging.file.dir=${DIR_BASE}\${DIR_TOMCAT}\logs;-Dcom.netiq.idm.osp.audit.enabled=false;"
$PARAM_TOMCAT_OPTS_COMMON_IDENT = "-Dlog4j.configuration=file:${DIR_BASE}\${DIR_TOMCAT}\conf\log4j.properties;-Dinternal.atlaslite.jcce.xml.w3c.XMLUtil.suppressSecurityWarning=true;"
$PARAM_TOMCAT_OPTS_GOV  = "-Dcom.netiq.ism.config.is.jndi.env=true;-Dnovell.logging.config.dir=${DIR_BASE}\${DIR_TOMCAT}\conf;"
$PARAM_TOMCAT_OPTS_RPT  = "-Dcom.netiq.rpt.config.file=${DIR_BASE}\${DIR_TOMCAT}\conf\ism-configuration.properties;-Dextend.local.config.dir=${DIR_BASE}\${DIR_TOMCAT}\conf;-Didmuserapp.logging.config.dir=${DIR_BASE}\${DIR_TOMCAT}\conf;"
$PARAM_TOMCAT_JVM_OPTS  = "${PARAM_TOMCAT_OPTS_COMMON_ALL}${PARAM_TOMCAT_OPTS_OSP}${PARAM_TOMCAT_OPTS_COMMON_IDENT}${PARAM_TOMCAT_OPTS_GOV}${PARAM_TOMCAT_OPTS_RPT}"
$PARAM_TOMCAT_JVM_MEM_INIT  = "1024"
$PARAM_TOMCAT_JVM_MEM_FINAL = "1024"

#Parse and consume input arguments provided
$USAGE_THIS = "Usage: " + $MyInvocation.MyCommand.Name + " -all|-gov|-base|-tomcat|-activeMQ|-postgreSQL|-java [<PG password>]"

$PARAM_INSTALL_ACTIVE_MQ = $false
$PARAM_INSTALL_JAVA = $false
$PARAM_INSTALL_POSTGRESQL = $false
$PARAM_INSTALL_WEB = $false

If ($help) {
    Write-Host "$USAGE_THIS"
    Exit
}
If ($All -or $Everything) {
    $PARAM_INSTALL_ACTIVE_MQ = $true
    $PARAM_INSTALL_JAVA = $true
    $PARAM_INSTALL_POSTGRESQL = $true
    $PARAM_INSTALL_WEB = $true
}
If ($Governance -or $Gov -or $IDGov) {
    $PARAM_INSTALL_ACTIVE_MQ = $true
    $PARAM_INSTALL_JAVA = $true
    $PARAM_INSTALL_WEB = $true
}
If ($Base -or $Tomcat -or $Web) {
    $PARAM_INSTALL_JAVA = $true
    $PARAM_INSTALL_WEB = $true
}
If ($ActiveMQ -or $Mail) {
    $PARAM_INSTALL_ACTIVE_MQ = $true
    $PARAM_INSTALL_JAVA = $true
}
If ($Java -or $JDK -or $JRE) {
    $PARAM_INSTALL_JAVA = $true
}
If ($Postgres -or $PostgreSQL -or $PG) {
    $PARAM_INSTALL_POSTGRESQL = $true
}

If (-not $WebState.equals("")) {
    $SERV_TOMCAT_STATE = $WebState
}

If (-not $Secret.equals("")) {
    $PARAM_POSTGRES_SECRET = $Secret
}

If (-not $PARAM_INSTALL_ACTIVE_MQ -and -not $PARAM_INSTALL_JAVA -and -not $PARAM_INSTALL_POSTGRESQL -and -not $PARAM_INSTALL_WEB) {
    Write-Host "Nothing specified for installation!"
    Write-Host "$USAGE_THIS"
    Exit
}
If ($PARAM_INSTALL_POSTGRESQL -and $PARAM_POSTGRES_SECRET -eq $null) {
    Write-Host "Required PostgreSQL password missing!"
    Write-Host "$USAGE_THIS"
    Exit
}

Function Unzip
{
    param([string]${ZIP_FILE}, [string]${OUT_DIR})
    [System.IO.Compression.ZipFile]::ExtractToDirectory(${ZIP_FILE}, ${OUT_DIR})
}

Function Enable-CatalinaOutput {
    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Enable Output to CATALINA_OUT:"
    Write-Host "######################################################################################"
    $text = (Get-Content -Path "${DIR_BASE}\${DIR_TOMCAT}\bin\catalina.bat" -ReadCount 0) -join "`r`n"
    $text -replace 'set _EXECJAVA=start "%TITLE%" %_RUNJAVA%', 'set _EXECJAVA=cmd /c "%_RUNJAVA%' | Set-Content -Path "${DIR_BASE}\${DIR_TOMCAT}\bin\catalina.bat"
    $text = (Get-Content -Path "${DIR_BASE}\${DIR_TOMCAT}\bin\catalina.bat" -ReadCount 0) -join "`r`n"
    $text -replace '%ACTION%', '%ACTION%" >> %CATALINA_OUT% 2>&1' | Set-Content -Path "${DIR_BASE}\${DIR_TOMCAT}\bin\catalina.bat"
}

Function Install-ActiveMQ {
    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Extracting ActiveMQ to ${DIR_BASE}\${DIR_ACTIVEMQ}:"
    Write-Host "######################################################################################"
    Unzip "${FILE_ACTIVEMQ}" "${DIR_BASE}"
    Rename-Item -Path "${DIR_BASE}\apache-activemq-${VSN_ACTIVEMQ}" -newName "${DIR_ACTIVEMQ}"
}

Function Install-JDK {
    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Extracting JDK to ${DIR_BASE}\${DIR_JDK}:"
    Write-Host "######################################################################################"
    Unzip "${FILE_JDK}" "${DIR_BASE}"
    Rename-Item -Path "${DIR_BASE}\zulu${VSN_JDK}-win_x64" -newName "${DIR_JDK}"

    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Creating symbolic link at ${DIR_BASE}\jre for ${DIR_JDK}\jre"
    Write-Host "######################################################################################"
    cmd /c "mklink /D ${DIR_BASE}\jre ${DIR_JDK}\jre"
}

Function Install-PostgreSQL {
    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Installing PostgreSQL to ${DIR_BASE}\${DIR_POSTGRES}:"
    Write-Host "######################################################################################"
    Start-Process ${FILE_POSTGRESQL} -Wait -NoNewWindow -ArgumentList `
        "--unattendedmodeui none", `
        "--mode unattended", `
        "--superaccount ${NAME_POSTGRES}", `
        "--superpassword ${PARAM_POSTGRES_SECRET}", `
        "--serverport ${PARAM_POSTGRES_PORT}", `
        "--prefix ${DIR_BASE}\${DIR_POSTGRES}", `
        "--datadir ${DIR_BASE}\${DIR_POSTGRES}\${DIR_POSTGRES_DATA}"

    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Updating PostgreSQL to allow remote access"
    Write-Host "######################################################################################"
    $text = (Get-Content -Path "${DIR_BASE}\${DIR_POSTGRES}\${DIR_POSTGRES_DATA}\pg_hba.conf" -ReadCount 0) -join "`n"
    $text -replace '/32', '/0 ' | Set-Content -Path "${DIR_BASE}\${DIR_POSTGRES}\${DIR_POSTGRES_DATA}\pg_hba.conf"

    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Restarting PostgreSQL"
    Write-Host "######################################################################################"
    Restart-Service -Name postgresql-x64-${VSN_POSTGRESQL_SERVICE}
}

Function Install-Tomcat {
    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Extracting Tomcat to ${DIR_BASE}\${DIR_TOMCAT}:"
    Write-Host "######################################################################################"
    Unzip "$FILE_TOMCAT" "${DIR_BASE}"
    Rename-Item -Path "${DIR_BASE}\apache-tomcat-${VSN_TOMCAT}" -newName "${DIR_TOMCAT}"

    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Moving the default Tomcat wars to ${DIR_BASE}\${DIR_BACKUP}:"
    Write-Host "######################################################################################"
    $null = (New-Item -Path "${DIR_BASE}" -Name "${DIR_BACKUP}" -ItemType "directory")
    Get-ChildItem -Path "${DIR_BASE}\${DIR_TOMCAT}\webapps" -Recurse | Move-Item -Destination "${DIR_BASE}\${DIR_BACKUP}"

    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Creating new default Tomcat index file at ${DIR_BASE}\${DIR_TOMCAT}\webapps\ROOT\index.html:"
    Write-Host "######################################################################################"
    $null = (New-Item -Path "${DIR_BASE}\${DIR_TOMCAT}\webapps" -Name "ROOT" -ItemType "directory")
    Copy-Item -Path "${DIR_PRISTINE}\index.html" -Destination "${DIR_BASE}\${DIR_TOMCAT}\webapps\ROOT\index.html"
}

Function Join-ActiveMQTomcat {
    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Copying the ActiveMQ jar to ${DIR_BASE}\${DIR_TOMCAT}\lib:"
    Write-Host "######################################################################################"
    Copy-Item "${DIR_BASE}\${DIR_ACTIVEMQ}\${DIR_ACTIVEMQ}-all*.jar" -Destination "${DIR_BASE}\${DIR_TOMCAT}\lib"
}

Function Setup-TomcatService {
    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Prepare ${DIR_BASE}\${DIR_TOMCAT}\bin\${NAME_TOMCAT_LAUNCH}:"
    Write-Host "######################################################################################"
    Copy-Item -Path "${DIR_PRISTINE}\${NAME_TOMCAT_LAUNCH}" -Destination "${DIR_BASE}\${DIR_TOMCAT}\bin\${NAME_TOMCAT_LAUNCH}"
    $text = (Get-Content -Path "${DIR_BASE}\${DIR_TOMCAT}\bin\${NAME_TOMCAT_LAUNCH}" -ReadCount 0) -join "`r`n"
    $text -replace "REPLACE_INSTALL_BASE", "${DIR_BASE}" | Set-Content -Path "${DIR_BASE}\${DIR_TOMCAT}\bin\${NAME_TOMCAT_LAUNCH}"
    $text = (Get-Content -Path "${DIR_BASE}\${DIR_TOMCAT}\bin\${NAME_TOMCAT_LAUNCH}" -ReadCount 0) -join "`r`n"
    $text -replace "REPLACE_INSTALL_JAVA", "${DIR_BASE}\${DIR_JDK}" | Set-Content -Path "${DIR_BASE}\${DIR_TOMCAT}\bin\${NAME_TOMCAT_LAUNCH}"

    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Prepare ${DIR_BASE}\${DIR_TOMCAT}\bin\${NAME_TOMCAT_QUIT}:"
    Write-Host "######################################################################################"
    Copy-Item -Path "${DIR_PRISTINE}\${NAME_TOMCAT_QUIT}" -Destination "${DIR_BASE}\${DIR_TOMCAT}\bin\${NAME_TOMCAT_QUIT}"
    $text = (Get-Content -Path "${DIR_BASE}\${DIR_TOMCAT}\bin\${NAME_TOMCAT_QUIT}" -ReadCount 0) -join "`r`n"
    $text -replace "REPLACE_INSTALL_BASE", "${DIR_BASE}" | Set-Content -Path "${DIR_BASE}\${DIR_TOMCAT}\bin\${NAME_TOMCAT_QUIT}"
    $text = (Get-Content -Path "${DIR_BASE}\${DIR_TOMCAT}\bin\${NAME_TOMCAT_QUIT}" -ReadCount 0) -join "`r`n"
    $text -replace "REPLACE_INSTALL_JAVA", "${DIR_BASE}\${DIR_JDK}" | Set-Content -Path "${DIR_BASE}\${DIR_TOMCAT}\bin\${NAME_TOMCAT_QUIT}"

    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Create the Tomcat service ${SERV_TOMCAT_SERVICE}:"
    Write-Host "######################################################################################"
    Copy-Item -Path "${DIR_PRISTINE}\${NAME_TOMCAT_SERVICE}" -Destination "${DIR_BASE}\${DIR_TOMCAT}\bin\${NAME_TOMCAT_SERVICE}"
    New-Service -Name "${SERV_TOMCAT_SERVICE}" -BinaryPathName "${DIR_BASE}\${DIR_TOMCAT}\bin\${NAME_TOMCAT_SERVICE} ${DIR_BASE}" -StartupType Manual
}

Function Restart-TomcatService {
    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Restart the Tomcat service, ${SERV_TOMCAT_SERVICE}:"
    Write-Host "######################################################################################"
    Restart-Service -Name "${SERV_TOMCAT_SERVICE}"
}

Function Start-TomcatService {
    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Start the Tomcat service, ${SERV_TOMCAT_SERVICE}:"
    Write-Host "######################################################################################"
    Start-Service -Name "${SERV_TOMCAT_SERVICE}"
}

Function Stop-TomcatService {
    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Stop the Tomcat service, ${SERV_TOMCAT_SERVICE}:"
    Write-Host "######################################################################################"
    Stop-Service -Name "${SERV_TOMCAT_SERVICE}"
}

Function Start-Recording {
    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Stopping existing transcripts"
    Write-Host "######################################################################################"
    $ErrorActionPreference="SilentlyContinue"
    Stop-Transcript | out-null

    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Creating log directory at: ${DIR_LOG}"
    Write-Host "######################################################################################"
    $null = (New-Item -Path "${DIR_LOG_BASE}" -Name "${DIR_LOG_PARENT}" -ItemType "directory")

    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Removing previous log if exists: ${FILE_LOG}"
    Write-Host "######################################################################################"
    Remove-Item -Path "${FILE_LOG}"

    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Starting transcript: ${FILE_LOG}"
    Write-Host "######################################################################################"
    $ErrorActionPreference = "Continue"
    Start-Transcript -path "${FILE_LOG}" -append
}

Function Stop-Recording {
    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Complete the transcript"
    Write-Host "######################################################################################"
    Stop-Transcript

    Write-Host ""
    Write-Host "######################################################################################"
    Write-Host " Completed"
    Write-Host " Review the log file: ${FILE_LOG}"
    Write-Host "######################################################################################"
}

Start-Recording
If ($PARAM_INSTALL_JAVA) { Install-JDK }
If ($PARAM_INSTALL_ACTIVE_MQ) { Install-ActiveMQ }
If ($PARAM_INSTALL_WEB) {
    Install-Tomcat
    Enable-CatalinaOutput
    If ($PARAM_INSTALL_ACTIVE_MQ) { Join-ActiveMQTomcat }
    Setup-TomcatService
    If ($SERV_TOMCAT_STATE.equals("start")) { Start-TomcatService }
    ElseIf ($SERV_TOMCAT_STATE.equals("restart")) { Restart-TomcatService }
    ElseIf ($SERV_TOMCAT_STATE.equals("stop")) { Stop-TomcatService }
}
if ($PARAM_INSTALL_POSTGRESQL) { Install-PostgreSQL }
Stop-Recording

# SIG # Begin signature block
# MIIggQYJKoZIhvcNAQcCoIIgcjCCIG4CAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQH8w7YFlLCE63JNLG
# KX7zUQIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCBJ61Ib/6I2qD99
# a7UN2uLA1Lei3OLPqaOSXL6I6I1uPKCCG4cwggPFMIICraADAgECAhACrFwmagtA
# m48LefKuRiV3MA0GCSqGSIb3DQEBBQUAMGwxCzAJBgNVBAYTAlVTMRUwEwYDVQQK
# EwxEaWdpQ2VydCBJbmMxGTAXBgNVBAsTEHd3dy5kaWdpY2VydC5jb20xKzApBgNV
# BAMTIkRpZ2lDZXJ0IEhpZ2ggQXNzdXJhbmNlIEVWIFJvb3QgQ0EwHhcNMDYxMTEw
# MDAwMDAwWhcNMzExMTEwMDAwMDAwWjBsMQswCQYDVQQGEwJVUzEVMBMGA1UEChMM
# RGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3d3cuZGlnaWNlcnQuY29tMSswKQYDVQQD
# EyJEaWdpQ2VydCBIaWdoIEFzc3VyYW5jZSBFViBSb290IENBMIIBIjANBgkqhkiG
# 9w0BAQEFAAOCAQ8AMIIBCgKCAQEAxszlc+b71LvlLS0ypt/lgT/JzSVJtnEqw9WU
# NGeiChywX2mmQLHEt7KP0JikqUFZOtPclNY823Q4pErMTSWC90qlUxI47vNJbXGR
# fmO2q6Zfw6SE+E9iUb74xezbOJLjBuUIkQzEKEFV+8taiRV+ceg1v01yCT2+OjhQ
# W3cxG42zxyRFmqesbQAUWgS3uhPrUQqYQUEiTmVhh4FBUKZ5XIneGUpX1S7mXRxT
# LH6YzRoGFqRoc9A0BBNcoXHTWnxV215k4TeHMFYE5RG0KYAS8Xk5iKICEXwnZreI
# t3jyygqoOKsKZMK/Zl2VhMGhJR6HXRpQCyASzEG7bgtROLhLywIDAQABo2MwYTAO
# BgNVHQ8BAf8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQUsT7DaQP4
# v0cB1JgmGggC72NkK8MwHwYDVR0jBBgwFoAUsT7DaQP4v0cB1JgmGggC72NkK8Mw
# DQYJKoZIhvcNAQEFBQADggEBABwaBpfc15yfPIhmBghXIdshR/gqZ6q/GDJ2QBBX
# wYrzetkRZY41+p78RbWe2UwxS7iR6EMsjrN4ztvjU3lx1uUhlAHaVYeaJGT2imbM
# 3pw3zag0sWmbI8ieeCIrcEPjVUcxYRnvWMWFL04w9qAxFiPI5+JlFjPLvxoboD34
# yl6LMYtgCIktDAZcUrfE+QqY0RVfnxK+fDZjOL1EpH/kJisKxJdpDemM4sAQV7jI
# dhKRVfJIadi8KgJbD0TUIDHb9LpwJl2QYJ68SxcJL7TLHkNoyQcnwdJc9+ohuWgS
# nDycv578gFybY83sR6olJ2egN/MAgn1U16n46S4To3foH0owggUoMIIEEKADAgEC
# AhANNvCcqoPZqIxHb9ZBXrLvMA0GCSqGSIb3DQEBCwUAMHYxCzAJBgNVBAYTAlVT
# MRUwEwYDVQQKEwxEaWdpQ2VydCBJbmMxGTAXBgNVBAsTEHd3dy5kaWdpY2VydC5j
# b20xNTAzBgNVBAMTLERpZ2lDZXJ0IFNIQTIgSGlnaCBBc3N1cmFuY2UgQ29kZSBT
# aWduaW5nIENBMB4XDTE4MTEzMDAwMDAwMFoXDTIxMTIwODEyMDAwMFowZzELMAkG
# A1UEBhMCVVMxDjAMBgNVBAgTBVRleGFzMRAwDgYDVQQHEwdIb3VzdG9uMRowGAYD
# VQQKExFOZXRJUSBDb3Jwb3JhdGlvbjEaMBgGA1UEAxMRTmV0SVEgQ29ycG9yYXRp
# b24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCeZBYzhJbJO6boi5Tm
# q+1hqQvSRI1lsJZx5CqI3uxWjgFpDyaGGaNeXCAYAcw/YXPNt+wyDMiRqWQ2bz33
# b/EkLDlNZaJUeJVSzG/JBlm8Wtgu+veqi9bph+pBR9cKOYz6cQG3GxD4SPU/zIhD
# wwt/M963pOaVkt4kl56jWWPjRQTnb71KWsyxP/c/Z5WsbSH+QOqur/JayO4LkH/i
# yHT44R+Xb5OK9/XAh3eFmYbFxgwudwx4kkpiS6ckneCzRE+SN7JyvdTS+lThopN8
# 5K41BgY2Edc+k0628AiSo/xuPpzimOn+Hxouax9MfXShWT6Hh+QNNFI/2h7QTS6Y
# t3uBAgMBAAGjggG/MIIBuzAfBgNVHSMEGDAWgBRnnQ8gCQzMijrlgkZyYvzxzJDl
# QDAdBgNVHQ4EFgQU4oPg/2aI0v0mcRoEwVgbZEHhuRYwDgYDVR0PAQH/BAQDAgeA
# MBMGA1UdJQQMMAoGCCsGAQUFBwMDMG0GA1UdHwRmMGQwMKAuoCyGKmh0dHA6Ly9j
# cmwzLmRpZ2ljZXJ0LmNvbS9zaGEyLWhhLWNzLWcxLmNybDAwoC6gLIYqaHR0cDov
# L2NybDQuZGlnaWNlcnQuY29tL3NoYTItaGEtY3MtZzEuY3JsMEwGA1UdIARFMEMw
# NwYJYIZIAYb9bAMLMCowKAYIKwYBBQUHAgEWHGh0dHBzOi8vd3d3LmRpZ2ljZXJ0
# LmNvbS9DUFMwCAYGZ4EMAQQBMIGIBggrBgEFBQcBAQR8MHowJAYIKwYBBQUHMAGG
# GGh0dHA6Ly9vY3NwLmRpZ2ljZXJ0LmNvbTBSBggrBgEFBQcwAoZGaHR0cDovL2Nh
# Y2VydHMuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0U0hBMkhpZ2hBc3N1cmFuY2VDb2Rl
# U2lnbmluZ0NBLmNydDAMBgNVHRMBAf8EAjAAMA0GCSqGSIb3DQEBCwUAA4IBAQBz
# dAg+QLbVAoX48OX0aZWGxUxBt/VCvwfWvyo0Z81E32M97dPS+z8rDPOQcD5cFmI/
# +GO3VAENqbAYuMN+OMq8qTqXU3EnmJekiF4rqIBSRBDUyYTrDvctyp/uM0DlmXcy
# sXEMEIa4t3UzEu0Ul/GAovRZVJ0ljQEAPsUP0JihcO7vakhfmP7B6AvKhYIUIJIR
# av7ML1aizJcKeV8lj1gDuSrG3KXqxGKnGSJnT1930YVJi8dmMvbx1AUmrEpiIQ6H
# LkKf1xAexMzAm/tpFp7W6vNfhST/UqkaJqOoR8FuhMmrCcCqjgKvlMjmMJaYkQ0X
# X/4fpBGqjWkNrFxnIuCAMIIFTzCCBDegAwIBAgIQC34QkDw4SQ/6L2eah6GnuTAN
# BgkqhkiG9w0BAQsFADBsMQswCQYDVQQGEwJVUzEVMBMGA1UEChMMRGlnaUNlcnQg
# SW5jMRkwFwYDVQQLExB3d3cuZGlnaWNlcnQuY29tMSswKQYDVQQDEyJEaWdpQ2Vy
# dCBIaWdoIEFzc3VyYW5jZSBFViBSb290IENBMB4XDTEzMTAyMjEyMDAwMFoXDTI4
# MTAyMjEyMDAwMFowdjELMAkGA1UEBhMCVVMxFTATBgNVBAoTDERpZ2lDZXJ0IElu
# YzEZMBcGA1UECxMQd3d3LmRpZ2ljZXJ0LmNvbTE1MDMGA1UEAxMsRGlnaUNlcnQg
# U0hBMiBIaWdoIEFzc3VyYW5jZSBDb2RlIFNpZ25pbmcgQ0EwggEiMA0GCSqGSIb3
# DQEBAQUAA4IBDwAwggEKAoIBAQC0Sl59Bw9B3sT1dhY2vXH/zz9Pc0uc0Q3+SstX
# WF6FFt0CFVSZ8I88L00CeBBoyNg1Sz/B92fOmByuM7ktHaQKVJPEhaLfNbH18Tyn
# szT7XUjJRsliRLxImesoSVPDPY/ADt41mOliUd89a0Bh7gRB2s+nXFaW0flMt0SE
# h5hp5YK5E+ZVv8iScJIKMW9/izKrz2tan2LEPu6+7VmkU38L8VKIinsKZyTLkM3s
# 0k00TLDhtZ+cxvZvLM3mylN0AZ9nNd44SS3O7TlEghl5Thqytfu7ePBJZqfP+lyW
# dZKLGnLZ/1CSU8w+wkMyCRqGE2k8+4EyMzJkdXMoJh0IMDsHAgMBAAGjggHhMIIB
# 3TASBgNVHRMBAf8ECDAGAQH/AgEAMA4GA1UdDwEB/wQEAwIBhjATBgNVHSUEDDAK
# BggrBgEFBQcDAzB/BggrBgEFBQcBAQRzMHEwJAYIKwYBBQUHMAGGGGh0dHA6Ly9v
# Y3NwLmRpZ2ljZXJ0LmNvbTBJBggrBgEFBQcwAoY9aHR0cDovL2NhY2VydHMuZGln
# aWNlcnQuY29tL0RpZ2lDZXJ0SGlnaEFzc3VyYW5jZUVWUm9vdENBLmNydDCBjwYD
# VR0fBIGHMIGEMECgPqA8hjpodHRwOi8vY3JsNC5kaWdpY2VydC5jb20vRGlnaUNl
# cnRIaWdoQXNzdXJhbmNlRVZSb290Q0EuY3JsMECgPqA8hjpodHRwOi8vY3JsMy5k
# aWdpY2VydC5jb20vRGlnaUNlcnRIaWdoQXNzdXJhbmNlRVZSb290Q0EuY3JsME8G
# A1UdIARIMEYwOAYKYIZIAYb9bAACBDAqMCgGCCsGAQUFBwIBFhxodHRwczovL3d3
# dy5kaWdpY2VydC5jb20vQ1BTMAoGCGCGSAGG/WwDMB0GA1UdDgQWBBRnnQ8gCQzM
# ijrlgkZyYvzxzJDlQDAfBgNVHSMEGDAWgBSxPsNpA/i/RwHUmCYaCALvY2QrwzAN
# BgkqhkiG9w0BAQsFAAOCAQEAag7/fhN8BqVLwC6M+VNkCeK6WJEwUOzMn+HTqC9I
# RjYYKdB4KF+YVkAPHrq9sTuHXNxb2CAN7RoWTdURJCFL8SdpkBPrEaEB2v21TnlZ
# db04KmrD9o5BK4qii9csUVHZnKDI4066bKhH0k7RaB+MAlc7sylqjmogKrnyAGJk
# usjpAPnMpNS6mjXYryxlbBZ8WCHeSjDQ+uskXQbJnRa3rUpF0yXiDPBAqlxNrH7N
# BoK5dkZpCNgytoL+46lYNEMbjmdnlz9oMRY2OJU+h/fHw6+dencZ2d6Ttf1uK/yU
# +T23TBI1LDC+6I2eBXCaSBP0jNbnHqw456jzrQy3euxn7TCCBmowggVSoAMCAQIC
# EAMBmgI6/1ixa9bV6uYX8GYwDQYJKoZIhvcNAQEFBQAwYjELMAkGA1UEBhMCVVMx
# FTATBgNVBAoTDERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2ljZXJ0LmNv
# bTEhMB8GA1UEAxMYRGlnaUNlcnQgQXNzdXJlZCBJRCBDQS0xMB4XDTE0MTAyMjAw
# MDAwMFoXDTI0MTAyMjAwMDAwMFowRzELMAkGA1UEBhMCVVMxETAPBgNVBAoTCERp
# Z2lDZXJ0MSUwIwYDVQQDExxEaWdpQ2VydCBUaW1lc3RhbXAgUmVzcG9uZGVyMIIB
# IjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAo2Rd/Hyz4II14OD2xirmSXU7
# zG7gU6mfH2RZ5nxrf2uMnVX4kuOe1VpjWwJJUNmDzm9m7t3LhelfpfnUh3SIRDsZ
# yeX1kZ/GFDmsJOqoSyyRicxeKPRktlC39RKzc5YKZ6O+YZ+u8/0SeHUOplsU/UUj
# joZEVX0YhgWMVYd5SEb3yg6Np95OX+Koti1ZAmGIYXIYaLm4fO7m5zQvMXeBMB+7
# NgGN7yfj95rwTDFkjePr+hmHqH7P7IwMNlt6wXq4eMfJBi5GEMiN6ARg27xzdPpO
# 2P6qQPGyznBGg+naQKFZOtkVCVeZVjCT88lhzNAIzGvsYkKRrALA76TwiRGPdwID
# AQABo4IDNTCCAzEwDgYDVR0PAQH/BAQDAgeAMAwGA1UdEwEB/wQCMAAwFgYDVR0l
# AQH/BAwwCgYIKwYBBQUHAwgwggG/BgNVHSAEggG2MIIBsjCCAaEGCWCGSAGG/WwH
# ATCCAZIwKAYIKwYBBQUHAgEWHGh0dHBzOi8vd3d3LmRpZ2ljZXJ0LmNvbS9DUFMw
# ggFkBggrBgEFBQcCAjCCAVYeggFSAEEAbgB5ACAAdQBzAGUAIABvAGYAIAB0AGgA
# aQBzACAAQwBlAHIAdABpAGYAaQBjAGEAdABlACAAYwBvAG4AcwB0AGkAdAB1AHQA
# ZQBzACAAYQBjAGMAZQBwAHQAYQBuAGMAZQAgAG8AZgAgAHQAaABlACAARABpAGcA
# aQBDAGUAcgB0ACAAQwBQAC8AQwBQAFMAIABhAG4AZAAgAHQAaABlACAAUgBlAGwA
# eQBpAG4AZwAgAFAAYQByAHQAeQAgAEEAZwByAGUAZQBtAGUAbgB0ACAAdwBoAGkA
# YwBoACAAbABpAG0AaQB0ACAAbABpAGEAYgBpAGwAaQB0AHkAIABhAG4AZAAgAGEA
# cgBlACAAaQBuAGMAbwByAHAAbwByAGEAdABlAGQAIABoAGUAcgBlAGkAbgAgAGIA
# eQAgAHIAZQBmAGUAcgBlAG4AYwBlAC4wCwYJYIZIAYb9bAMVMB8GA1UdIwQYMBaA
# FBUAEisTmLKZB+0e36K+Vw0rZwLNMB0GA1UdDgQWBBRhWk0ktkkynUoqeRqDS/Qe
# icHKfTB9BgNVHR8EdjB0MDigNqA0hjJodHRwOi8vY3JsMy5kaWdpY2VydC5jb20v
# RGlnaUNlcnRBc3N1cmVkSURDQS0xLmNybDA4oDagNIYyaHR0cDovL2NybDQuZGln
# aWNlcnQuY29tL0RpZ2lDZXJ0QXNzdXJlZElEQ0EtMS5jcmwwdwYIKwYBBQUHAQEE
# azBpMCQGCCsGAQUFBzABhhhodHRwOi8vb2NzcC5kaWdpY2VydC5jb20wQQYIKwYB
# BQUHMAKGNWh0dHA6Ly9jYWNlcnRzLmRpZ2ljZXJ0LmNvbS9EaWdpQ2VydEFzc3Vy
# ZWRJRENBLTEuY3J0MA0GCSqGSIb3DQEBBQUAA4IBAQCdJX4bM02yJoFcm4bOIyAP
# gIfliP//sdRqLDHtOhcZcRfNqRu8WhY5AJ3jbITkWkD73gYBjDf6m7GdJH7+IKRX
# rVu3mrBgJuppVyFdNC8fcbCDlBkFazWQEKB7l8f2P+fiEUGmvWLZ8Cc9OB0obzpS
# CfDscGLTYkuw4HOmksDTjjHYL+NtFxMG7uQDthSr849Dp3GdId0UyhVdkkHa+Q+B
# 0Zl0DSbEDn8btfWg8cZ3BigV6diT5VUW8LsKqxzbXEgnZsijiwoc5ZXarsQuWaBh
# 3drzbaJh6YoLbewSGL33VVRAA5Ira8JRwgpIr7DUbuD0FAo6G+OPPcqvao173NhE
# MIIGzTCCBbWgAwIBAgIQBv35A5YDreoACus/J7u6GzANBgkqhkiG9w0BAQUFADBl
# MQswCQYDVQQGEwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3
# d3cuZGlnaWNlcnQuY29tMSQwIgYDVQQDExtEaWdpQ2VydCBBc3N1cmVkIElEIFJv
# b3QgQ0EwHhcNMDYxMTEwMDAwMDAwWhcNMjExMTEwMDAwMDAwWjBiMQswCQYDVQQG
# EwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3d3cuZGlnaWNl
# cnQuY29tMSEwHwYDVQQDExhEaWdpQ2VydCBBc3N1cmVkIElEIENBLTEwggEiMA0G
# CSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDogi2Z+crCQpWlgHNAcNKeVlRcqcTS
# QQaPyTP8TUWRXIGf7Syc+BZZ3561JBXCmLm0d0ncicQK2q/LXmvtrbBxMevPOkAM
# Rk2T7It6NggDqww0/hhJgv7HxzFIgHweog+SDlDJxofrNj/YMMP/pvf7os1vcyP+
# rFYFkPAyIRaJxnCI+QWXfaPHQ90C6Ds97bFBo+0/vtuVSMTuHrPyvAwrmdDGXRJC
# geGDboJzPyZLFJCuWWYKxI2+0s4Grq2Eb0iEm09AufFM8q+Y+/bOQF1c9qjxL6/s
# iSLyaxhlscFzrdfx2M8eCnRcQrhofrfVdwonVnwPYqQ/MhRglf0HBKIJAgMBAAGj
# ggN6MIIDdjAOBgNVHQ8BAf8EBAMCAYYwOwYDVR0lBDQwMgYIKwYBBQUHAwEGCCsG
# AQUFBwMCBggrBgEFBQcDAwYIKwYBBQUHAwQGCCsGAQUFBwMIMIIB0gYDVR0gBIIB
# yTCCAcUwggG0BgpghkgBhv1sAAEEMIIBpDA6BggrBgEFBQcCARYuaHR0cDovL3d3
# dy5kaWdpY2VydC5jb20vc3NsLWNwcy1yZXBvc2l0b3J5Lmh0bTCCAWQGCCsGAQUF
# BwICMIIBVh6CAVIAQQBuAHkAIAB1AHMAZQAgAG8AZgAgAHQAaABpAHMAIABDAGUA
# cgB0AGkAZgBpAGMAYQB0AGUAIABjAG8AbgBzAHQAaQB0AHUAdABlAHMAIABhAGMA
# YwBlAHAAdABhAG4AYwBlACAAbwBmACAAdABoAGUAIABEAGkAZwBpAEMAZQByAHQA
# IABDAFAALwBDAFAAUwAgAGEAbgBkACAAdABoAGUAIABSAGUAbAB5AGkAbgBnACAA
# UABhAHIAdAB5ACAAQQBnAHIAZQBlAG0AZQBuAHQAIAB3AGgAaQBjAGgAIABsAGkA
# bQBpAHQAIABsAGkAYQBiAGkAbABpAHQAeQAgAGEAbgBkACAAYQByAGUAIABpAG4A
# YwBvAHIAcABvAHIAYQB0AGUAZAAgAGgAZQByAGUAaQBuACAAYgB5ACAAcgBlAGYA
# ZQByAGUAbgBjAGUALjALBglghkgBhv1sAxUwEgYDVR0TAQH/BAgwBgEB/wIBADB5
# BggrBgEFBQcBAQRtMGswJAYIKwYBBQUHMAGGGGh0dHA6Ly9vY3NwLmRpZ2ljZXJ0
# LmNvbTBDBggrBgEFBQcwAoY3aHR0cDovL2NhY2VydHMuZGlnaWNlcnQuY29tL0Rp
# Z2lDZXJ0QXNzdXJlZElEUm9vdENBLmNydDCBgQYDVR0fBHoweDA6oDigNoY0aHR0
# cDovL2NybDMuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0QXNzdXJlZElEUm9vdENBLmNy
# bDA6oDigNoY0aHR0cDovL2NybDQuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0QXNzdXJl
# ZElEUm9vdENBLmNybDAdBgNVHQ4EFgQUFQASKxOYspkH7R7for5XDStnAs0wHwYD
# VR0jBBgwFoAUReuir/SSy4IxLVGLp6chnfNtyA8wDQYJKoZIhvcNAQEFBQADggEB
# AEZQPsm3KCSnOB22WymvUs9S6TFHq1Zce9UNC0Gz7+x1H3Q48rJcYaKclcNQ5IK5
# I9G6OoZyrTh4rHVdFxc0ckeFlFbR67s2hHfMJKXzBBlVqefj56tizfuLLZDCwNK1
# lL1eT7EF0g49GqkUW6aGMWKoqDPkmzmnxPXOHXh2lCVz5Cqrz5x2S+1fwksW5Etw
# TACJHvzFebxMElf+X+EevAJdqP77BzhPDcZdkbkPZ0XN1oPt55INjbFpjE/7WeAj
# D9KqrgB87pxCDs+R1ye3Fu4Pw718CqDuLAhVhSK46xgaTfwqIa1JMYNHlXdx3LEb
# S0scEJx3FMGdTy9alQgpECYxggRQMIIETAIBATCBijB2MQswCQYDVQQGEwJVUzEV
# MBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3d3cuZGlnaWNlcnQuY29t
# MTUwMwYDVQQDEyxEaWdpQ2VydCBTSEEyIEhpZ2ggQXNzdXJhbmNlIENvZGUgU2ln
# bmluZyBDQQIQDTbwnKqD2aiMR2/WQV6y7zANBglghkgBZQMEAgEFAKCBhDAYBgor
# BgEEAYI3AgEMMQowCKACgAChAoAAMBkGCSqGSIb3DQEJAzEMBgorBgEEAYI3AgEE
# MBwGCisGAQQBgjcCAQsxDjAMBgorBgEEAYI3AgEVMC8GCSqGSIb3DQEJBDEiBCDR
# z6IrBB7qmoTpm4nRu+c1PbpeSt/G5wSvejGyC47zmTANBgkqhkiG9w0BAQEFAASC
# AQAgQg5p09MlMMsPCLNxfPNN1ZUbRdY7o6lfZHbfWJgoNZ6gqaZOvNBGzruXBRij
# 40Dgldk+8Wj24+xG6Rb6LTmRC2OY/lRrYP+FS7bS2EfHJ3u9fhgm/1Px5cWcXqF7
# bz6luP6pvWvCY33NOLPni42JdFd11iR5kIFLBSKZE2y/TcC2TDUDklipjXfbVbBo
# IVEegqwyQUBT3/ff3mEHIj1wBhQQ5/qWTwmRY2yj68dEF22Z/OuGKiOhInmin5zK
# 37dGm9OHkrkJlYrDPU1eWHcPZmOermNl6aFUJJUb7LIjM43yN7USMv0kGLpiZZY5
# I/d7i7BSr9ljWFp7rca18f7DoYICDzCCAgsGCSqGSIb3DQEJBjGCAfwwggH4AgEB
# MHYwYjELMAkGA1UEBhMCVVMxFTATBgNVBAoTDERpZ2lDZXJ0IEluYzEZMBcGA1UE
# CxMQd3d3LmRpZ2ljZXJ0LmNvbTEhMB8GA1UEAxMYRGlnaUNlcnQgQXNzdXJlZCBJ
# RCBDQS0xAhADAZoCOv9YsWvW1ermF/BmMAkGBSsOAwIaBQCgXTAYBgkqhkiG9w0B
# CQMxCwYJKoZIhvcNAQcBMBwGCSqGSIb3DQEJBTEPFw0xOTEyMDQxODQ4MTdaMCMG
# CSqGSIb3DQEJBDEWBBTPh6lrm81sf+u1KOFa8kEx4oT9OzANBgkqhkiG9w0BAQEF
# AASCAQCZ1c1rzJaJqRRKejrRGnjMfMXsDNkr2MZXZxgFhk6b14KNR2CrHbvc3UGC
# 8jo/Nw9G0lX60sHeFJMNeeRz2V5ezkV6xICZ6/Iz42uj3u2+9JNiQ1mt4gnBDMgI
# WkWZsOv3zR5N5o8u+bl67NUpKXanAuou4m/m6rdLGGU9HC+/sc1l9XISX/CFOlUt
# ylNBKLDEPDbNBs1oHUeU6WGpzSakB1OXPx50ZL5wKGUZKlpcZiwA3pNOHuEW/eUd
# eSR1zJwk4Svpdd+jXmHl9I6LfVe+uxWpzZSTPrvVLb9c2+1TZdxXbvdzFa0O/wzr
# gLfSXyASWUswUBcsrKU6+Ut1waWj
# SIG # End signature block
